#define _CRT_SECURE_NO_WARNINGS
#define WIN32_LEAN_AND_MEAN 

#include <windows.h> 
#include <stdint.h>
#include <stdlib.h>
#include "small_dll.h"     
#include "mcs48.h"   

extern "C" IDSIMMODEL __declspec(dllexport) *createdsimmodel(CHAR *device, ILICENCESERVER *ils) {
	ils->authorize(0x80808048);
    return new mcs48;
}

extern "C" void __declspec(dllexport)deletedsimmodel(mcs48 *model) {
    delete model;
}

mcs48::~mcs48() {
    if (ram != NULL) free(ram);
    if (rom != NULL) free(rom);
}

void mcs48::setup(IINSTANCE *inst, IDSIMCKT *dsim) {
	ckt = dsim;
    iinst = inst;
    pin_reset = inst->getdsimpin("$RESET$", true);
	pin_reset->sethandler(this, (PINHANDLERFN)&mcs48::pin_reset_onchange);
    pin_ss = inst->getdsimpin("$SS$", true);
	pin_ss->sethandler(this, (PINHANDLERFN)&mcs48::pin_ss_onchange);
    pin_ea = inst->getdsimpin("EA", true);
	pin_ea->sethandler(this, (PINHANDLERFN)&mcs48::pin_ea_onchange);
	pin_int = inst->getdsimpin("$INT$", true);
	pin_int->sethandler(this, (PINHANDLERFN)&mcs48::pin_int_onchange);
    pin_t0 = inst->getdsimpin("T0", true);
    pin_t0->drivetristate(0);
	pin_t0->sethandler(this, (PINHANDLERFN)&mcs48::pin_t0_onchange);
    pin_t1 = inst->getdsimpin("T1", true);
	pin_t1->sethandler(this, (PINHANDLERFN)&mcs48::pin_t1_onchange);
    pin_ale = inst->getdsimpin("ALE", true);
    pin_ale->drivebool(0, 0);
    pin_prog = inst->getdsimpin("$PROG$", true);
    pin_prog->drivebool(0, 1);
    pin_rd = inst->getdsimpin("$RD$", true);
    pin_rd->drivebool(0, 1);
    pin_psen = inst->getdsimpin("$PSEN$", true);
    pin_psen->drivebool(0, 1);
    pin_wr = inst->getdsimpin("$WR$", true);
    pin_wr->drivebool(0, 1);
    pin_x1 = inst->getdsimpin("XTAL1", true);
    pin_x1->sethandler(this, (PINHANDLERFN)&mcs48::pin_x1_onchange);
	char pstr[8] = "DB0"; 
    unsigned i;
    for (i = BUS; i < BUS + 8; i++) {
        pin_ports[i] = inst->getdsimpin(pstr, true);
        pin_ports[i]->drivetristate(0);
        pstr[2]++;
    }
	lstrcpyA(pstr, "P1.0");
    for (i = P1; i < P1 + 8; i++) {
        pin_ports[i] = inst->getdsimpin(pstr, true);
        pin_ports[i]->drivestate(0, WHI);
        pstr[3]++;
    }
	lstrcpyA(pstr, "P2.0");
    for (i = P2; i < P2 + 8; i++) {
        pin_ports[i] = inst->getdsimpin(pstr, true);
        pin_ports[i]->drivestate(0, WHI);
        pstr[3]++;
    }
    DWORD d = inst->getinitval("CLK");
    if (d) ckt->setclockcallback(0, 1500000000000 / d, this, (CALLBACKHANDLERFN)&mcs48::callback, 0);
    else opt |= O_EXT_CLK;
    if (inst->getinitval("NOALE") == 0) cmd |= C_ALE;
    else inst->log("mcs48: suppress ALE enabled.");
    DWORD rm = inst->getinitval("RAM", 0);
    if ((rm == 64) || (rm == 128) || (rm == 256)) {
        rammask = uint8_t(rm - 1);
        inst->log("mcs48: internal RAM size set to %u bytes.", rm);
    }
    char* fname = inst->getstrval("FILE");
    if (fname != NULL) {
        HANDLE f = CreateFileA(fname, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, 0);
        DWORD sz = 0;
        if (f != INVALID_HANDLE_VALUE) {
            sz = GetFileSize(f, NULL);
            if ((sz != 0) && (sz != INVALID_FILE_SIZE)) {
                if (sz <= 1024) romsize = 1024;
                else if (sz <= 2048) romsize = 2048;
                else if (sz <= 4096) romsize = 4096;
                else romsize = 8192;
                rom = (uint8_t*)malloc(romsize);
				memset(rom, 0x66, romsize);
                sz = _lread((HFILE)f, rom, romsize);
                if ((sz != 0) && (sz != INVALID_FILE_SIZE)) {
                    inst->log("mcs48: ROM file '%s' loaded %u bytes. ROM size set to %u bytes.", fname, sz, romsize);
                }
            }
            CloseHandle(f);
        }
        if ((sz == 0) || (sz == INVALID_FILE_SIZE)) {
            inst->error("mcs48: unable to load ROM file '%s'   ERROR: %u.", fname, GetLastError());
            return;
        }
    }
	if (romsize == 0) cmd |= C_PSEN;
    ram = (uint8_t*)malloc(rammask + 1);
}

BOOL mcs48::indicate(REALTIME time, ACTIVEDATA *data) {
    return false;
}

int mcs48::isdigital(CHAR *pinname) {
    return 1;
}

void mcs48::simulate(ABSTIME time, DSIMMODES mode) {
}

#pragma optimize("y", on)
uint8_t mcs48::get_port_data(unsigned port, uint8_t mask) {
    uint8_t tmp = 0;
	uint8_t b = 1;
	for (unsigned i = 0; i < 8; i++) {
		if ((mask & b) && (!islow(pin_ports[port + i]->istate()))) tmp |= b;
		b += b;
	}
    return tmp;
}

void mcs48::set_port_data(ABSTIME time, uint8_t data, unsigned port, uint8_t mask) {
	uint8_t b = 1;
	for (unsigned i = 0; i < 8; i++) {
		if (mask & b) pin_ports[port + i]->drivebool(time, data & b);
		b += b;
	}
}

void mcs48::set_port_tristate(ABSTIME time, unsigned port, uint8_t mask) {
	uint8_t b = 1;
	for (unsigned i = 0; i < 8; i++) {
		if (mask & b) pin_ports[port + i]->drivetristate(time);
		b += b;
	}
}

void mcs48::set_port_data_w(ABSTIME time, uint8_t data, unsigned port, uint8_t mask) {
	uint8_t b = 1;
	for (unsigned i = 0; i < 8; i++) {
		if (mask & b) pin_ports[port + i]->drivestate(time, data & b ? WHI : SLO);
		b += b;
	}
}

#define flag(_f) (((_f) & psw) ? 1 : 0)
#define setflag(_f) {psw |= (_f);}
#define clrflag(_f) {psw &= ~(_f);}
#define pushs(_d) {ram[sp++] = (_d); if (sp > 23) sp = 8;}
#define pushsl(_d) {ram[sp++] = (_d);}
#define pop() (sp--, (sp < 8) ? (sp = 23) : 0, ram[sp])
#define popl() (ram[--sp])
#define opreg ((psw & FBS ? 0b11000 : 0) + (opcode & 7))
#define m2_req() {cmd |= C_M2REQ;}

void mcs48::callback(ABSTIME time, EVENTID eventid) {
	if (eventid != 0) return;
	if (eflag & EOUT_CLK) pin_t0->drivebool(time, cc & 1);
	switch (cc++) {
	case 0:
		return;
	case 1:
		if (cmd & (C_ALE | C_BUSA)) pin_ale->drivebool(time, 1);
		if (cmd & C_BUSA) set_port_data(time, busaddr, BUS, PALL);
		return;
	case 2:
		if (cmd & (C_ERD | C_EWR)) set_port_data(time, extcmd, P2, PLO);
		return;
	case 3:
		if (cmd & (C_ALE | C_BUSA)) pin_ale->drivebool(time, 0);
		if (cmd & (C_ERD | C_EWR)) pin_prog->drivebool(time, 0);
		if (cmd & C_ERD) set_port_tristate(time + 70000, P2, PLO);
		return;
	case 4:
		if (cmd & C_EWR) set_port_data(time, a, P2, PLO);
		else if (cmd & C_RD) set_port_tristate(time, BUS, PALL);
		return;
	case 5:
		if (cmd & C_WR) {
			set_port_data(time, busdata, BUS, PALL);
			pin_wr->drivebool(time, 0);
		}
		else if (cmd & C_RD) {
			pin_rd->drivebool(time, 0);
		}
		return;
	case 6:
		if (ecmd == P1SETA) {
			set_port_data_w(time, p1data, P1, PALL);
			ecmd = 0;
		}
		else if (ecmd == P2SETA) {
			set_port_data_w(time, p2data, P2, PALL);
			ecmd = 0;
		}
		return;
	case 7:
		if (cmd & C_ERD) {
			a = get_port_data(P2, PLO);
		}
		goto L17;
	case 8:
		if (cmd & C_WR) {
			pin_wr->drivebool(time, 1);
		}
		else if (cmd & C_RD) {
			a = get_port_data(BUS, PALL);
		}
		else if (cmd & (C_ERD | C_EWR)) {
			pin_prog->drivebool(time, 1);
		}
		return;
	case 9:
		if (cmd & C_WR) {
			if (cmd & C_BUSA) set_port_tristate(time, BUS, PALL);
		}
		else if (cmd & C_RD) {
			pin_rd->drivebool(time, 1);
		}
		else if (cmd & (C_ERD | C_EWR)) {
			set_port_data_w(time, p2data, P2, PLO);
		}
		return;
	case 10:
		if (ecmd == INVALID) {
			ecmd = 0;
			iinst->warning("mcs48: invalid opcode 0x%02X at 0x%04X.", opcode, pctmp);
		}
		if (opt & O_RESET) {
			cc = 10;
			return;
		}
		if (((itflag & IACT) == 0) && ((cmd & C_M2REQ) == 0)) {
			if ((opt & O_INT) && (itflag & IEN)) {
				itflag |= IACT;
				eflag &= ~EHALT;
				pushsl(pc & 0xFF);
				pushs(((pc & 0xF00) >> 8) | psw);
				pc = 3;
				return;
			}
			if ((itflag & TOVF) && (itflag & IENTCNT)) {
				itflag &= ~TOVF;
				itflag |= IACT;
				eflag &= ~EHALT;
				pushsl(pc & 0xFF);
				pushs(((pc & 0xF00) >> 8) | psw);
				pc = 7;
				return;
			}
		}
		if (eflag & EHALT) {
			cc = 9;
			return;
		}
		return;
	case 11:
		if (cmd & (C_ALE | C_PSEN))pin_ale->drivebool(time, 1);
		if (cmd & C_PSEN) set_port_data(time, pc & 0xFF, BUS, PALL);
		return;
	case 12:
		if (cmd & C_PSEN) set_port_data(time, pc >> 8, P2, PLO);
		return;
	case 13:
		if (cmd & (C_ALE | C_PSEN))pin_ale->drivebool(time, 0);
		return;
	case 14:
		if (cmd & C_PSEN) {
			set_port_tristate(time, BUS, PALL);
			pin_psen->drivebool(time, 0);
		}
		return;
	case 15:
		return;
	case 16:
		if (ecmd == BUSSET) {
			set_port_data_w(time, busdata, BUS, pmask);
			ecmd = 0;
		}
		else if (ecmd == P1SETD) {
			set_port_data_w(time, p1data, P1, pmask);
			ecmd = 0;
		}
		else if (ecmd == P2SETD) {
			set_port_data_w(time, p2data, P2, pmask);
			ecmd = 0;
		}
		else if (ecmd == P1GET) {
			a = get_port_data(P1, PALL);
			ecmd = 0;
		}
		else if (ecmd == P2GET) {
			a = get_port_data(P2, PALL);
			ecmd = 0;
		}
		return;
	case 17:
	L17:	
		if (opt & O_SS) {
			cc -= 2;
			return;
		}
		if (itflag & TRUN) {
			prescaler++;
			prescaler &= 31;
			if (prescaler == 0) {
				timer++;
				if (timer == 0) itflag |= TOVF;
			}
		}
		else if (itflag & TCNT) {
			if (itflag & TT1) {
				itflag &= ~TT1;
				timer++;
				if (timer == 0) itflag |= TOVF;
			}
		}
		return;
	case 18:
		if (cmd & C_PSEN) {
			romdata = get_port_data(BUS, PALL);
			pin_psen->drivebool(time, 1);
			set_port_data_w(time, p2data, P2, PLO);
		}
		else {
			romdata = rom[pc | pcadd];
		}
		if ((cmd & C_M2REQ) == 0) {
			if (romdata == 0x66) {
				ckt->suspend(iinst, NULL);
			}
		}
		return;
	case 19:
		cc = 10;
		uint8_t arg;
		bool m2 = false;
		if (cmd & C_M2REQ) {
			m2 = true;
			arg = romdata;
		}
		else {
			opcode = romdata;
		}
		cmd &= C_ALE;
		uint16_t ppc = pc;
		pc = (ppc & 0x800) | ((ppc + 1) & 0x7FF);
		uint16_t tmp16;
		uint8_t tmp8;
		switch (opcode) {
		case 0x03:	// ADD A,#imm
		case 0x13:	// ADDC A,#imm
			if (m2) {
				tmp16 = 0;
				if (opcode & 0x10) tmp16 = flag(FC);
				clrflag(FC | FAC);
				if (((a & 0x0F) + (arg & 0x0F) + tmp16) > 0x0F) setflag(FAC);
				tmp16 += a + arg;
				if (tmp16 > 0xFF) setflag(FC);
				a = (tmp16 & 0xFF);
				break;
			}
			m2_req();
			break;
		case 0x68: case 0x69: case 0x6A: case 0x6B: case 0x6C: case 0x6D: case 0x6E: case 0x6F:	// ADD A,Rn
		case 0x78: case 0x79: case 0x7A: case 0x7B: case 0x7C: case 0x7D: case 0x7E: case 0x7F:	// ADDC A,Rn
		case 0x60: case 0x61:	// ADD A,@Rn
		case 0x70: case 0x71:	// ADDC A,@Rn
			arg = ram[opreg];
			if ((opcode & 8) == 0) arg = ram[rammask & arg];
			tmp16 = 0;
			if (opcode & 0x10) tmp16 = flag(FC);
			clrflag(FC | FAC);
			if (((a & 0x0F) + (arg & 0x0F) + tmp16) > 0x0F) setflag(FAC);
			tmp16 += a + arg;
			if (tmp16 > 0xFF) setflag(FC);
			a = (tmp16 & 0xFF);
			break;
		case 0x53:	// ANL A,#imm
			if (m2) {
				a &= arg;
				break;
			}
			m2_req();
			break;
		case 0x58: case 0x59: case 0x5A: case 0x5B:	case 0x5C: case 0x5D: case 0x5E: case 0x5F:	// ANL A,Rn
			a &= ram[opreg];
			break;
		case 0x50: case 0x51: // ANL A,@Rn
			a &= ram[rammask & ram[opreg]];
			break;
		case 0x98:	// ANL BUS,#imm
			if (m2) {
				busdata &= arg;
				ecmd = BUSSET;
				pmask = ~arg;
				break;
			}
			m2_req();
			break;
		case 0x99: // ANL P1,#imm
			if (m2) {
				p1data &= arg;
				ecmd = P1SETD;
				pmask = ~arg;
				break;
			}
			m2_req();
			break;
		case 0x9A:	// ANL P2,#imm 
			if (m2) {
				p2data &= arg;
				ecmd = P2SETD;
				pmask = ~arg;
				break;
			}
			m2_req();
			break;
		case 0x9C: case 0x9D: case 0x9E: case 0x9F:	// ANLD Pp,A 
			extcmd = (opcode & 3) | ECMD_AND;
			cmd |= C_EWR;
			break;
		case 0x14: case 0x34: case 0x54: case 0x74: case 0x94: case 0xB4: case 0xD4: case 0xF4: // CALL
			if (m2) {
				pushsl(pc & 0xFF);
				pushs(((pc & 0xF00) >> 8) | psw);
				pc = arg | (((uint16_t)opcode << 3) & 0x700) | (((eflag & EROMBANK) && ((itflag & IACT) == 0)) ? 0x800 : 0);
				break;
			}
			m2_req();
			break;
		case 0x27:	// CLR A
			a = 0;
			break;
		case 0x97:	// CLR C
			clrflag(FC);
			break;
		case 0xA5:	// CLR F1
			eflag &= ~EF1;
			break;
		case 0x85:	// CLR F0
			clrflag(FF0);
			break;
		case 0x37:	// CPL A
			a ^= 0xFF;
			break;
		case 0xA7:	// CPL C
			psw ^= FC;
			break;
		case 0x95:	// CPL F0
			psw ^= FF0;
			break;
		case 0xB5:	// CPL F1
			eflag ^= EF1;
			break;
		case 0x57:	// DA A
			if (((a & 0x0F) > 0x09) || flag(FAC)) {
				a += 6;
				if ((a & 0xF0) == 0) setflag(FC);
			}
			if ((a & 0xF0) > 0x90 || (flag(FC))) {
				a += 0x60;
				setflag(FC);
			}
			break;
		case 0x07:	// DEC A
			a--;
			break;
		case 0xC8: case 0xC9: case 0xCA: case 0xCB: case 0xCC: case 0xCD: case 0xCE: case 0xCF:	// DEC Rn
			ram[opreg]--;
			break;
		case 0x15:	// DIS I
			itflag &= ~IEN;
			break;
		case 0x35:	// DIS TCNTI
			itflag &= ~IENTCNT;
			break;
		case 0xE8: case 0xE9: case 0xEA: case 0xEB: case 0xEC: case 0xED: case 0xEE: case 0xEF:	// DJNZ Rn,addr
			if (m2) {
				tmp8 = --ram[opreg];
				if (tmp8) {
					pc = (ppc & 0xF00) | arg;
				}
				break;
			}
			m2_req();
			break;
		case 0x05:	// EN I
			itflag |= IEN;
			break;
		case 0x25:	// EN TCNTI
			itflag |= IENTCNT;
			break;
		case 0x75:	// ENT0 CLK
			eflag |= EOUT_CLK;
			break;
		case 0x01:	// HLT
			eflag |= EHALT;
			break;
		case 0x08:	// INS A,BUS
			cmd |= C_RD;
			break;
		case 0x09: // IN A,P1
			ecmd = P1GET;
			cc = 0;
			break;
		case 0x0A: // IN A,P2
			ecmd = P2GET;
			cc = 0;
			break;	
		case 0x17: // INC A
			a++;
			break;
		case 0x18: case 0x19: case 0x1A: case 0x1B: case 0x1C: case 0x1D: case 0x1E: case 0x1F:	// INC Rn
			ram[opreg]++;
			break;
		case 0x10: case 0x11:	// INC @Rn
			ram[rammask & ram[opreg]]++;
			break;
		case 0x12: case 0x32: case 0x52: case 0x72: case 0x92: case 0xB2: case 0xD2: case 0xF2: // JBn addr
			if (m2) {
				if (a & _BV(opcode >> 5)) pc = (ppc & 0xF00) | arg;
				break;
			}
			m2_req();
			break;
		case 0xF6:		// JC addr
			if (m2) {
				if (flag(FC)) pc = (ppc & 0xF00) | arg;
				break;
			}
			m2_req();
			break;
		case 0xB6:		// JF0 addr
			if (m2) {
				if (flag(FF0)) pc = (ppc & 0xF00) | arg;
				break;
			}
			m2_req();
			break;
		case 0x76:		// JF1 addr
			if (m2) {
				if (eflag & EF1) pc = (ppc & 0xF00) | arg;
				break;
			}
			m2_req();
			break;
		case 0x04: case 0x24: case 0x44: case 0x64: case 0x84: case 0xA4: case 0xC4: case 0xE4: // JMP addr
			if (m2) {
				pc = arg | (((uint16_t)opcode << 3) & 0x700) | (((eflag & EROMBANK) && ((itflag & IACT) == 0)) ? 0x800 : 0);
				break;
			}
			m2_req();
			break;
		case 0xB3:		// JMPP @A
			if (m2) {
				pc = (pctmp & 0xF00) | arg;
				break;
			}
			pctmp = ppc;
			pc = (ppc & 0xF00) | a;
			m2_req();
			break;
		case 0xE6:		// JNC addr
			if (m2) {
				if (flag(FC) == 0) pc = (ppc & 0xF00) | arg;
				break;
			}
			m2_req();
			break;
		case 0x86:		// JNI addr
			if (m2) {
				if (opt & O_INT) pc = (ppc & 0xF00) | arg;
				break;
			}
			m2_req();
			break;		
		case 0x26:		// JNT0 addr
			if (m2) {
				if ((opt & O_T0) == 0) pc = (ppc & 0xF00) | arg;
				break;
			}
			m2_req();
			break;
		case 0x36:		// JT0 addr
			if (m2) {
				if (opt & O_T0) pc = (ppc & 0xF00) | arg;
				break;
			}
			m2_req();
			break;
		case 0x46:		// JNT1 addr
			if (m2) {
				if ((opt & O_T1) == 0) pc = (ppc & 0xF00) | arg;
				break;
			}
			m2_req();
			break;
		case 0x56:		// JT1 addr
			if (m2) {
				if (opt & O_T1) pc = (ppc & 0xF00) | arg;
				break;
			}
			m2_req();
			break;
		case 0x96:		// JNZ addr
			if (m2) {
				if (a) pc = (ppc & 0xF00) | arg;
				break;
			}
			m2_req();
			break;
		case 0xC6:		// JZ addr
			if (m2) {
				if (a == 0) pc = (ppc & 0xF00) | arg;
				break;
			}
			m2_req();
			break;
		case 0x16:		// JTF addr
			if (m2) {
				if (itflag & TOVF) pc = (ppc & 0xF00) | arg;
				itflag &= ~TOVF;
				break;
			}
			m2_req();
			break;
		case 0x23:		// MOV A,#imm
			if (m2) {
				a = arg;
				break;
			}
			m2_req();
			break;
		case 0xF8: case 0xF9: case 0xFA: case 0xFB: case 0xFC: case 0xFD: case 0xFE: case 0xFF:	// MOV A,Rn
			a = ram[opreg];
			break;
		case 0xF0: case 0xF1:	// MOV A,@Rn
			a = ram[rammask & ram[opreg]];
			break;
		case 0xC7:		// MOV A,PSW
			a = psw | ((sp - 8) >> 1) | 8;
			break;
		case 0x42:		// MOV A,T
			a = timer;
			break;
		case 0x62:		// MOV T,A
			timer = a;
			break;
		case 0xD7:		// MOV PSW,A
			psw = (a & 0xF0);
			sp = ((a & 7) << 1) + 8;
			break;
		case 0xA8: case 0xA9: case 0xAA: case 0xAB: case 0xAC: case 0xAD: case 0xAE: case 0xAF:	// MOV Rn,A
			ram[opreg] = a;
			break;
		case 0xB8: case 0xB9: case 0xBA: case 0xBB: case 0xBC: case 0xBD: case 0xBE: case 0xBF:	// MOV Rn,#imm
			if (m2) {
				ram[opreg] = arg;
				break;
			}
			m2_req();
			break;
		case 0xA0: case 0xA1: // MOV @Rn,A
			ram[rammask & ram[opreg]] = a;
			break;
		case 0xB0: case 0xB1: // MOV @Rn,#imm
			if (m2) {
				ram[rammask & ram[opreg]] = arg;
				break;
			}
			m2_req();
			break;
		case 0x0C: case 0x0D: case 0x0E: case 0x0F:	// MOVD A,Pp
			extcmd = opcode & 3;
			cmd |= C_ERD;
			break;
		case 0x3C: case 0x3D: case 0x3E: case 0x3F:	// MOVD Pp,A 
			extcmd = (opcode & 3) | ECMD_WR;
			cmd |= C_EWR;
			break;
		case 0xA3:	// MOVP A,@A
			if (m2) {
				pc = pctmp;
				a = arg;
				break;
			}
			pctmp = pc;
			pc = (ppc & 0xF00) | a;
			m2_req();
			break;
		case 0xE3:	// MOVP3 A,@A 
			if (m2) {
				pc = pctmp;
				a = arg;
				break;
			}
			pctmp = pc;
			pc = 0x300 | a;
			m2_req();
			break;
		case 0x80: case 0x81:	// MOVX A,@Rn
			busaddr = ram[opreg];
			cmd |= C_BUSA | C_RD;
			break;
		case 0x90: case 0x91:	// MOVX @Rn,A
			busaddr = ram[opreg];
			busdata = a;
			cmd |= C_BUSA | C_WR;
			break;
		case 0x00:	// NOP
			break;
		case 0x43:	// ORL A,#imm
			if (m2) {
				a |= arg;
				break;
			}
			m2_req();
			break;
		case 0x48: case 0x49: case 0x4A: case 0x4B: case 0x4C: case 0x4D: case 0x4E: case 0x4F:	// ORL A,Rn
			a |= ram[opreg];
			break;
		case 0x40: case 0x41:	// ORL A,@Rn
			a |= ram[rammask & ram[opreg]];
			break;
		case 0x88:	// ORL BUS,#imm
			if (m2) {
				busdata |= arg;
				ecmd = BUSSET;
				pmask = arg;
				break;
			}
			m2_req();
			break;
		case 0x89:  // ORL P1,#imm
			if (m2) {
				p1data |= arg;
				ecmd = P1SETD;
				pmask = arg;
				break;
			}
			m2_req();
			break;
		case 0x8A:	// ORL P2,#imm 
			if (m2) {
				p2data |= arg;
				ecmd = P2SETD;
				pmask = arg;
				break;
			}
			m2_req();
			break;
		case 0x8C: case 0x8D: case 0x8E: case 0x8F:	// ORLD Pp,A 
			extcmd = (opcode & 3) | ECMD_OR;
			cmd |= C_EWR;
			break;
		case 0x02:	// OUTL BUS,A
			busdata = a;
			cmd |= C_WR;
			break;
		case 0x39:  // OUTL P1,A
			p1data = a;
			ecmd = P1SETA;
			cc = 0;
			break;
		case 0x3A:	// OUTL P2,A
			p2data = a;
			ecmd = P2SETA;
			cc = 0;
			break;
		case 0x93:	// RETR
			itflag &= ~IACT;
			arg = pop();
			psw = arg & 0xF0;
			pc = ((arg & 0x0F) << 8);
			pc |= popl();
			cc = 0;
			break;
		case 0x83:	// RET
			arg = pop();
			pc = ((arg & 0x0F) << 8);
			pc |= popl();
			cc = 0;
			break;
		case 0xE7:	// RL A
			a = ((a << 1) & 0xFE) | ((a >> 7) & 0x01);
			break;
		case 0xF7:	// RLC A
			arg = ((a << 1) & 0xFE) | flag(FC);
			clrflag(FC);
			if (a & 0x80) setflag(FC);
			a = arg;
			break;
		case 0x77:	// RR A
			a = ((a >> 1) & 0x7F) | ((a << 7) & 0x80);
			break;
		case 0x67:	// RRC A
			arg = ((a >> 1) & 0x7F) | flag(FC) * 0x80;
			clrflag(FC);
			if (a & 0x01) setflag(FC);
			a = arg;
			break;
		case 0xE5:	// SEL MB0
			eflag &= ~EROMBANK; 
			break;
		case 0xF5:	// SEL MB1
			eflag |= EROMBANK;
			break;
		case 0xC5:	// SEL RB0
			clrflag(FBS);
			break;
		case 0xD5:	// SEL RB1
			setflag(FBS);
			break;
		case 0x65:	// STOP TCNT
			itflag &= ~(TCNT | TRUN);
			break;
		case 0x45:	// STRT CNT
			itflag &= ~(TRUN & TT1);
			itflag |= TCNT;
			break;
		case 0x55:	// STRT T
			itflag &= ~TCNT;
			itflag |= TRUN;
			break;
		case 0x47:	// SWAP A
			a = ((a << 4) & 0xF0) | ((a >> 4) & 0x0F);
			break;
		case 0x28: case 0x29: case 0x2A: case 0x2B: case 0x2C: case 0x2D: case 0x2E: case 0x2F:	// XCH A,Rn
			arg = ram[opreg];
			ram[opreg] = a;
			a = arg;
			break;
		case 0x20: case 0x21: // XCH A,@Rn
			tmp8 = rammask & ram[opreg];
			arg = ram[tmp8];
			ram[tmp8] = a;
			a = arg;
			break;
		case 0x30: case 0x31: // XCHD A,@Rn
			tmp8 = rammask & ram[opreg];
			arg = ram[tmp8];
			ram[tmp8] = (ram[tmp8] & 0xF0) | (a & 0x0F);
			a = (a & 0xF0) | (arg & 0x0F);
			break;
		case 0xD3:	// XRL A,#imm
			if (m2) {
				a ^= arg;
				break;
			}
			m2_req();
			break;
		case 0xD8: case 0xD9: case 0xDA: case 0xDB: case 0xDC: case 0xDD: case 0xDE: case 0xDF:	// XRL A,Rn
			a ^= ram[opreg];
			break;
		case 0xD0: case 0xD1:	// XRL A,@Rn
			a ^= ram[rammask & ram[opreg]];
			break;
		default: // invalid opcode
			ecmd = INVALID;
			pctmp = ppc;
			break;
		}
		if (cmd & (C_RD | C_WR | C_ERD | C_EWR | C_BUSA)) cc = 0;
		if (pc >= romsize) cmd |= C_PSEN;
	}
}

void mcs48::pin_x1_onchange(ABSTIME time, DSIMMODES mode) {
    if (pin_x1->isposedge()) {
        if (opt & O_EXT_CLK) {
            cx1++;
            if (cx1 == 3) cx1 = 0;
            if (cx1) callback(time, 0);
            return;
        }
        if ((opt & O_EXT_CLK_WARN) == 0) {
            opt |= O_EXT_CLK_WARN;
            iinst->warning("mcs48: 'CLK' property is set, and 'XTAL1' pin changed.");
        }
    }
}

void mcs48::pin_reset_onchange(ABSTIME time, DSIMMODES mode) {
	if (pin_reset->isinactive()) {
		opt |= O_RESET;
		pin_t0->drivetristate(time);
		pin_prog->drivebool(time, 1);
		pin_rd->drivebool(time, 1);
		pin_wr->drivebool(time, 1);
		pin_psen->drivebool(time, 1);
		pin_ale->drivebool(time, 0);
		set_port_tristate(time, BUS, PALL);
		set_port_data_w(time, 0xFF, P1, PALL);
		set_port_data_w(time, 0xFF, P2, PALL);
		cmd &= C_ALE;
		if (romsize == 0) cmd |= C_PSEN;
		pc = 0;
		cc = 10;
		busdata = 0xFF;
		p1data = 0xFF;
		p2data = 0xFF;
		ecmd = 0;
		a = 0;
		psw = 0;
		itflag = 0;
		timer = 0;
		eflag = 0;
		sp = 8;
		prescaler = 0;
	}
	else {
		opt &= ~O_RESET;
	}
}

void mcs48::pin_int_onchange(ABSTIME time, DSIMMODES mode) {
	if (pin_int->isinactive()) opt |= O_INT; else opt &= ~O_INT;
}

void mcs48::pin_t0_onchange(ABSTIME time, DSIMMODES mode) {
	if (eflag & EOUT_CLK) return;
	if (pin_t0->isinactive()) opt &= ~O_T0; else opt |= O_T0;
}

void mcs48::pin_t1_onchange(ABSTIME time, DSIMMODES mode) {
	if (pin_t1->isinactive()) {
		opt &= ~O_T1;
		itflag |= TT1;
	}
	else opt |= O_T1;
}

void mcs48::pin_ss_onchange(ABSTIME time, DSIMMODES mode) {
	if (pin_ss->isinactive()) opt |= O_SS; else opt &= ~O_SS;
}

void mcs48::pin_ea_onchange(ABSTIME time, DSIMMODES mode) {
	if (romsize < 0x1FFF) return;
	if (pin_ea->isactive()) pcadd = 0x1000; else pcadd = 0;
}

void mcs48::runctrl(RUNMODES mode) {
}

void mcs48::actuate(REALTIME time, ACTIVESTATE newstate) {
}

