#pragma once
#include "vsm.hpp" 


#define BUS     0
#define P1      8
#define P2      16
#define PLO     0x0F
#define PALL    0xFF
#define _BV(bit) (1 << (bit))

#define C_ALE   1
#define C_PSEN  2
#define C_RD    4
#define C_WR	8
#define C_ERD	16
#define C_EWR	32
#define C_BUSA  64
#define C_M2REQ	128

#define ECMD_WR     0b0100
#define ECMD_OR     0b1000
#define ECMD_AND    0b1100

#define O_EXT_CLK       1
#define O_EXT_CLK_WARN  2
#define O_INT   4
#define O_RESET 8
#define O_T0    16
#define O_T1    32
#define O_SS    64

#define P1SETA  1
#define P2SETA  2
#define P1GET   3
#define P2GET   4
#define BUSSET  5
#define P1SETD  6
#define P2SETD  7
#define INVALID 8
                
#define FC      128
#define FAC     64
#define FF0     32
#define FBS     16

#define IEN     1
#define IACT    2
#define IENTCNT 4
#define TCNT    8
#define TRUN    16
#define TOVF    32
#define TT1     64

#define EROMBANK    1
#define EOUT_CLK    2
#define EF1         4
#define EHALT       8

class mcs48 : public IDSIMMODEL {
public:
    int isdigital(CHAR* pinname);
    void setup(IINSTANCE* inst, IDSIMCKT* dsim);
    void runctrl(RUNMODES mode);
    void actuate(REALTIME time, ACTIVESTATE newstate);
    BOOL indicate(REALTIME time, ACTIVEDATA* data);
    void simulate(ABSTIME time, DSIMMODES mode);
    void callback(ABSTIME time, EVENTID eventid);
    ~mcs48();
private:
    void pin_x1_onchange(ABSTIME time, DSIMMODES mode);
    void pin_reset_onchange(ABSTIME time, DSIMMODES mode);
    void pin_int_onchange(ABSTIME time, DSIMMODES mode);
    void pin_t0_onchange(ABSTIME time, DSIMMODES mode);
    void pin_t1_onchange(ABSTIME time, DSIMMODES mode);
    void pin_ea_onchange(ABSTIME time, DSIMMODES mode);
    void pin_ss_onchange(ABSTIME time, DSIMMODES mode);
    uint8_t _fastcall get_port_data(unsigned port, uint8_t mask);
    void set_port_data(ABSTIME time, uint8_t data, unsigned port, uint8_t mask);
    void set_port_tristate(ABSTIME time, unsigned port, uint8_t mask);
    void set_port_data_w(ABSTIME time, uint8_t data, unsigned port, uint8_t mask);
    IDSIMPIN* pin_reset;
    IDSIMPIN* pin_ss;
    IDSIMPIN* pin_ea;
    IDSIMPIN* pin_int;
    IDSIMPIN* pin_t0;
    IDSIMPIN* pin_t1;
    IDSIMPIN* pin_ale;
    IDSIMPIN* pin_prog;
    IDSIMPIN* pin_rd;
    IDSIMPIN* pin_psen;
    IDSIMPIN* pin_wr;
    IDSIMPIN* pin_x1;
    IDSIMPIN* pin_ports[24];
    IINSTANCE* iinst;
    IDSIMCKT* ckt;
    uint8_t* ram = NULL;
    uint8_t* rom = NULL;
    uint16_t romsize = 0;
    uint16_t pcadd = 0;
    uint16_t pctmp = 0;
    uint16_t pc = 0;
    uint8_t cc = 10;
    uint8_t busdata = 0xFF;
    uint8_t p1data = 0xFF;
    uint8_t p2data = 0xFF;
    uint8_t cmd = 0;
    uint8_t ecmd = 0;
    uint8_t a = 0;
    uint8_t psw = 0;
    uint8_t itflag = 0;
    uint8_t timer = 0;
    uint8_t eflag = 0;
    uint8_t sp = 8;
    uint8_t prescaler = 0;
    uint8_t rammask = 0xFF;
    uint8_t cx1 = 0xFF;
    uint8_t opt = 0;
    uint8_t extcmd = 0;
    uint8_t busaddr = 0;
    uint8_t romdata = 0;
    uint8_t opcode = 0;
    uint8_t pmask = 0;
};